@@ -32,7 +32,6 @@
 #include <unistd.h>
 
 #include <bluetooth/bluetooth.h>
-#include <bluetooth/l2cap.h>
 #include <bluetooth/hci.h>
 #include <bluetooth/hci_lib.h>
 #include <bluetooth/sdp.h>
@@ -146,76 +145,30 @@ static GIOChannel *do_connect(gboolean le)
 	return chan;
 }
 
-static void primary_all_cb(guint8 status, const guint8 *pdu, guint16 plen,
-							gpointer user_data)
+static void primary_all_cb(GSList *services, guint8 status, gpointer user_data)
 {
-	GAttrib *attrib = user_data;
-	struct att_data_list *list;
-	unsigned int i;
-	uint16_t end;
+	GSList *l;
 
-	if (status == ATT_ECODE_ATTR_NOT_FOUND)
-		goto done;
-
-	if (status != 0) {
+	if (status) {
 		g_printerr("Discover all primary services failed: %s\n",
 							att_ecode2str(status));
 		goto done;
 	}
 
-	list = dec_read_by_grp_resp(pdu, plen);
-	if (list == NULL)
-		goto done;
-
-	for (i = 0, end = 0; i < list->num; i++) {
-		char uuidstr[MAX_LEN_UUID_STR];
-		uint8_t *value = list->data[i];
-		uint8_t length;
-		uint16_t start;
-		uuid_t uuid;
-
-		/* Each element contains: attribute handle, end group handle
-		 * and attribute value */
-		length = list->len - 2 * sizeof(uint16_t);
-		start = att_get_u16(value);
-		end = att_get_u16(&value[2]);
-
-		g_print("attr handle = 0x%04x, end grp handle = 0x%04x, ",
-								start, end);
-		if (length == 2)
-			sdp_uuid16_create(&uuid, att_get_u16(&value[4]));
-		else
-			sdp_uuid128_create(&uuid, value + 4);
-
-		sdp_uuid2strn(&uuid, uuidstr, MAX_LEN_UUID_STR);
-		g_print("attr value (UUID) = %s\n", uuidstr);
+	for (l = services; l; l = l->next) {
+		struct att_primary *prim = l->data;
+		g_print("attr handle = 0x%04x, end grp handle = 0x%04x "
+			"uuid: %s\n", prim->start, prim->end, prim->uuid);
 	}
 
-	att_data_list_free(list);
-
-	/* Don't go beyond the maximum handle value */
-	if (end == 0xffff)
-		goto done;
-
-	/*
-	 * Discover all primary services sub-procedure shall send another
-	 * Read by Group Type Request until Error Response is received and
-	 * the Error Code is set to Attribute Not Found.
-	 */
-
-	gatt_discover_primary(attrib, end + 1, opt_end, NULL, primary_all_cb,
-								attrib);
-	return;
-
 done:
-	if (opt_listen == FALSE)
-		g_main_loop_quit(event_loop);
+	g_main_loop_quit(event_loop);
 }
 
-static void primary_by_uuid_cb(guint8 status, const guint8 *pdu, guint16 plen,
+static void primary_by_uuid_cb(GSList *ranges, guint8 status,
 							gpointer user_data)
 {
-	GSList *ranges, *l;
+	GSList *l;
 
 	if (status != 0) {
 		g_printerr("Discover primary services by UUID failed: %s\n",
@@ -223,21 +176,12 @@ static void primary_by_uuid_cb(guint8 status, const guint8 *pdu, guint16 plen,
 		goto done;
 	}
 
-	ranges = dec_find_by_type_resp(pdu, plen);
-	if (ranges == NULL) {
-		g_printerr("Protocol error!\n");
-		goto done;
-	}
-
 	for (l = ranges; l; l = l->next) {
 		struct att_range *range = l->data;
 		g_print("Starting handle: %04x Ending handle: %04x\n",
 						range->start, range->end);
 	}
 
-	g_slist_foreach(ranges, (GFunc) g_free, NULL);
-	g_slist_free(ranges);
-
 done:
 	g_main_loop_quit(event_loop);
 }
@@ -273,7 +217,7 @@ static void events_handler(const uint8_t *pdu, uint16_t len, gpointer user_data)
 	olen = enc_confirmation(opdu, sizeof(opdu));
 
 	if (olen > 0)
-		g_attrib_send(attrib, opdu[0], opdu, olen, NULL, NULL, NULL);
+		g_attrib_send(attrib, 0, opdu[0], opdu, olen, NULL, NULL, NULL);
 }
 
 static gboolean listen_start(gpointer user_data)
@@ -293,11 +237,10 @@ static gboolean primary(gpointer user_data)
 	GAttrib *attrib = user_data;
 
 	if (opt_uuid)
-		gatt_discover_primary(attrib, opt_start, opt_end, opt_uuid,
-						primary_by_uuid_cb, attrib);
+		gatt_discover_primary(attrib, opt_uuid, primary_by_uuid_cb,
+									NULL);
 	else
-		gatt_discover_primary(attrib, opt_start, opt_end, NULL,
-						primary_all_cb, attrib);
+		gatt_discover_primary(attrib, NULL, primary_all_cb, NULL);
 
 	return FALSE;
 }
@@ -401,9 +344,14 @@ done:
 static void char_read_by_uuid_cb(guint8 status, const guint8 *pdu,
 					guint16 plen, gpointer user_data)
 {
+	struct characteristic_data *char_data = user_data;
 	struct att_data_list *list;
 	int i;
 
+	if (status == ATT_ECODE_ATTR_NOT_FOUND &&
+					char_data->start != opt_start)
+		goto done;
+
 	if (status != 0) {
 		g_printerr("Read characteristics by UUID failed: %s\n",
 							att_ecode2str(status));
@@ -418,6 +366,8 @@ static void char_read_by_uuid_cb(guint8 status, const guint8 *pdu,
 		uint8_t *value = list->data[i];
 		int j;
 
+		char_data->start = att_get_u16(value) + 1;
+
 		g_print("handle: 0x%04x \t value: ", att_get_u16(value));
 		value += 2;
 		for (j = 0; j < list->len - 2; j++, value++)
@@ -427,7 +377,14 @@ static void char_read_by_uuid_cb(guint8 status, const guint8 *pdu,
 
 	att_data_list_free(list);
 
+	gatt_read_char_by_uuid(char_data->attrib, char_data->start,
+					char_data->end, opt_uuid,
+					char_read_by_uuid_cb,
+					char_data);
+
+	return;
 done:
+	g_free(char_data);
 	g_main_loop_quit(event_loop);
 }
 
@@ -436,8 +393,16 @@ static gboolean characteristics_read(gpointer user_data)
 	GAttrib *attrib = user_data;
 
 	if (opt_uuid != NULL) {
+		struct characteristic_data *char_data;
+
+		char_data = g_new(struct characteristic_data, 1);
+		char_data->attrib = attrib;
+		char_data->start = opt_start;
+		char_data->end = opt_end;
+
 		gatt_read_char_by_uuid(attrib, opt_start, opt_end, opt_uuid,
-						char_read_by_uuid_cb, attrib);
+						char_read_by_uuid_cb, char_data);
+
 		return FALSE;
 	}
 
