@@ -37,6 +37,8 @@
 
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/hci.h>
+#include <bluetooth/sdp.h>
+#include <bluetooth/sdp_lib.h>
 #include <bluetooth/mgmt.h>
 
 #include "plugin.h"
@@ -60,6 +62,7 @@ static struct controller_info {
 	uint8_t hci_ver;
 	uint16_t hci_rev;
 	gboolean enabled;
+	gboolean connectable;
 	gboolean discoverable;
 	gboolean pairable;
 	uint8_t sec_mode;
@@ -87,7 +90,7 @@ static void read_version_complete(int sk, void *buf, size_t len)
 	DBG("version %u revision %u", mgmt_version, mgmt_revision);
 
 	memset(&hdr, 0, sizeof(hdr));
-	hdr.opcode = MGMT_OP_READ_INDEX_LIST;
+	hdr.opcode = htobs(MGMT_OP_READ_INDEX_LIST);
 	if (write(sk, &hdr, sizeof(hdr)) < 0)
 		error("Unable to read controller index list: %s (%d)",
 						strerror(errno), errno);
@@ -115,7 +118,7 @@ static void read_info(int sk, uint16_t index)
 	struct mgmt_cp_read_info *cp = (void *) &buf[sizeof(*hdr)];
 
 	memset(buf, 0, sizeof(buf));
-	hdr->opcode = MGMT_OP_READ_INFO;
+	hdr->opcode = htobs(MGMT_OP_READ_INFO);
 	hdr->len = htobs(sizeof(*cp));
 
 	cp->index = htobs(index);
@@ -171,10 +174,104 @@ static void mgmt_index_removed(int sk, void *buf, size_t len)
 	remove_controller(index);
 }
 
+static int mgmt_set_mode(int index, uint16_t opcode, uint8_t val)
+{
+	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_mode)];
+	struct mgmt_hdr *hdr = (void *) buf;
+	struct mgmt_mode *cp = (void *) &buf[sizeof(*hdr)];
+
+	memset(buf, 0, sizeof(buf));
+	hdr->opcode = htobs(opcode);
+	hdr->len = htobs(sizeof(*cp));
+
+	cp->index = htobs(index);
+	cp->val = val;
+
+	if (write(mgmt_sock, buf, sizeof(buf)) < 0)
+		return -errno;
+
+	return 0;
+}
+
+static int mgmt_set_connectable(int index, gboolean connectable)
+{
+	DBG("index %d connectable %d", index, connectable);
+	return mgmt_set_mode(index, MGMT_OP_SET_CONNECTABLE, connectable);
+}
+
+static int mgmt_set_discoverable(int index, gboolean discoverable)
+{
+	DBG("index %d discoverable %d", index, discoverable);
+	return mgmt_set_mode(index, MGMT_OP_SET_DISCOVERABLE, discoverable);
+}
+
+static int mgmt_set_pairable(int index, gboolean pairable)
+{
+	DBG("index %d pairable %d", index, pairable);
+	return mgmt_set_mode(index, MGMT_OP_SET_PAIRABLE, pairable);
+}
+
+static int mgmt_update_powered(int index, uint8_t powered)
+{
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	gboolean pairable, discoverable;
+	uint8_t on_mode;
+
+	if (index > max_index) {
+		error("Unexpected index %u", index);
+		return -ENODEV;
+	}
+
+	info = &controllers[index];
+
+	info->enabled = powered;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (adapter == NULL) {
+		DBG("Adapter not found");
+		return -ENODEV;
+	}
+
+	if (!powered) {
+		info->connectable = FALSE;
+		info->pairable = FALSE;
+		info->discoverable = FALSE;
+
+		btd_adapter_stop(adapter);
+		return 0;
+	}
+
+	btd_adapter_start(adapter);
+
+	btd_adapter_get_mode(adapter, NULL, &on_mode, &pairable);
+
+	discoverable = (on_mode == MODE_DISCOVERABLE);
+
+	if (on_mode == MODE_DISCOVERABLE && !info->discoverable)
+		mgmt_set_discoverable(index, TRUE);
+	else if (on_mode == MODE_CONNECTABLE && !info->connectable)
+		mgmt_set_connectable(index, TRUE);
+	else {
+		uint8_t mode = 0;
+
+		if (info->connectable)
+			mode |= SCAN_PAGE;
+		if (info->discoverable)
+			mode |= SCAN_INQUIRY;
+
+		adapter_mode_changed(adapter, mode);
+	}
+
+	if (info->pairable != pairable)
+		mgmt_set_pairable(index, pairable);
+
+	return 0;
+}
+
 static void mgmt_powered(int sk, void *buf, size_t len)
 {
-	struct btd_adapter *adapter;
-	struct mgmt_ev_powered *ev = buf;
+	struct mgmt_mode *ev = buf;
 	uint16_t index;
 
 	if (len < sizeof(*ev)) {
@@ -184,25 +281,165 @@ static void mgmt_powered(int sk, void *buf, size_t len)
 
 	index = btohs(bt_get_unaligned(&ev->index));
 
+	DBG("Controller %u powered %u", index, ev->val);
+
+	mgmt_update_powered(index, ev->val);
+}
+
+static void mgmt_discoverable(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *ev = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+	uint8_t mode;
+
+	if (len < sizeof(*ev)) {
+		error("Too small discoverable event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	DBG("Controller %u discoverable %u", index, ev->val);
+
 	if (index > max_index) {
-		DBG("Ignoring powered event for unknown controller %u", index);
+		error("Unexpected index %u in discoverable event", index);
 		return;
 	}
 
-	controllers[index].enabled = ev->powered;
+	info = &controllers[index];
 
-	DBG("Controller %u powered %s", index, ev->powered ? "on" : "off");
+	info->discoverable = ev->val ? TRUE : FALSE;
 
-	adapter = manager_find_adapter(&controllers[index].bdaddr);
-	if (adapter == NULL) {
-		DBG("Adapter not found");
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
 		return;
-	}
 
-	if (ev->powered)
-		btd_adapter_start(adapter);
+	if (info->connectable)
+		mode = SCAN_PAGE;
 	else
-		btd_adapter_stop(adapter);
+		mode = 0;
+
+	if (info->discoverable)
+		mode |= SCAN_INQUIRY;
+
+	adapter_mode_changed(adapter, mode);
+}
+
+static void mgmt_connectable(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *ev = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+	uint8_t mode;
+
+	if (len < sizeof(*ev)) {
+		error("Too small connectable event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	DBG("Controller %u connectable %u", index, ev->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in connectable event", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->connectable = ev->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
+		return;
+
+	if (info->discoverable)
+		mode = SCAN_INQUIRY;
+	else
+		mode = 0;
+
+	if (info->connectable)
+		mode |= SCAN_PAGE;
+
+	adapter_mode_changed(adapter, mode);
+}
+
+static void mgmt_pairable(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *ev = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+
+	if (len < sizeof(*ev)) {
+		error("Too small pairable event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	DBG("Controller %u pairable %u", index, ev->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in pairable event", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->pairable = ev->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
+		return;
+
+	btd_adapter_pairable_changed(adapter, info->pairable);
+}
+
+static void uuid_to_uuid128(uuid_t *uuid128, const uuid_t *uuid)
+{
+	if (uuid->type == SDP_UUID16)
+		sdp_uuid16_to_uuid128(uuid128, uuid);
+	else if (uuid->type == SDP_UUID32)
+		sdp_uuid32_to_uuid128(uuid128, uuid);
+	else
+		memcpy(uuid128, uuid, sizeof(*uuid));
+}
+
+static int mgmt_uuid_op(int index, uint16_t op, uuid_t *uuid)
+{
+	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_cp_uuid)];
+	struct mgmt_hdr *hdr = (void *) buf;
+	struct mgmt_cp_uuid *cp = (void *) &buf[sizeof(*hdr)];
+	uuid_t uuid128;
+
+	uuid_to_uuid128(&uuid128, uuid);
+
+	memset(buf, 0, sizeof(buf));
+	hdr->opcode = htobs(op);
+	hdr->len = htobs(sizeof(*cp));
+
+	cp->index = htobs(index);
+	memcpy(cp->uuid, uuid128.value.uuid128.data, 16);
+
+	if (write(mgmt_sock, buf, sizeof(buf)) < 0)
+		return -errno;
+
+	return 0;
+}
+
+static int clear_uuids(int index)
+{
+	uuid_t uuid_any;
+
+	memset(&uuid_any, 0, sizeof(uuid_any));
+	uuid_any.type = SDP_UUID128;
+
+	return mgmt_uuid_op(index, MGMT_OP_REMOVE_UUID, &uuid_any);
 }
 
 static void read_index_list_complete(int sk, void *buf, size_t len)
@@ -230,40 +467,14 @@ static void read_index_list_complete(int sk, void *buf, size_t len)
 
 		add_controller(index);
 		read_info(sk, index);
+		clear_uuids(index);
 	}
 }
 
-static int mgmt_set_discoverable(int index, gboolean discoverable)
-{
-	DBG("index %d discoverable %d", index, discoverable);
-	return -ENOSYS;
-}
-
-static int mgmt_set_pairable(int index, gboolean pairable)
-{
-	DBG("index %d pairable %d", index, pairable);
-	return -ENOSYS;
-}
-
 static int mgmt_set_powered(int index, gboolean powered)
 {
-	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_cp_set_powered)];
-	struct mgmt_hdr *hdr = (void *) buf;
-	struct mgmt_cp_set_powered *cp = (void *) &buf[sizeof(*hdr)];
-
 	DBG("index %d powered %d", index, powered);
-
-	memset(buf, 0, sizeof(buf));
-	hdr->opcode = MGMT_OP_SET_POWERED;
-	hdr->len = htobs(sizeof(*cp));
-
-	cp->index = htobs(index);
-	cp->powered = powered;
-
-	if (write(mgmt_sock, buf, sizeof(buf)) < 0)
-		return -errno;
-
-	return 0;
+	return mgmt_set_mode(index, MGMT_OP_SET_POWERED, powered);
 }
 
 static void read_info_complete(int sk, void *buf, size_t len)
@@ -272,7 +483,6 @@ static void read_info_complete(int sk, void *buf, size_t len)
 	struct controller_info *info;
 	struct btd_adapter *adapter;
 	uint8_t mode;
-	gboolean pairable, discoverable;
 	uint16_t index;
 	char addr[18];
 
@@ -290,6 +500,7 @@ static void read_info_complete(int sk, void *buf, size_t len)
 	info = &controllers[index];
 	info->type = rp->type;
 	info->enabled = rp->powered;
+	info->connectable = rp->connectable;
 	info->discoverable = rp->discoverable;
 	info->pairable = rp->pairable;
 	info->sec_mode = rp->sec_mode;
@@ -316,26 +527,138 @@ static void read_info_complete(int sk, void *buf, size_t len)
 		return;
 	}
 
-	btd_adapter_get_mode(adapter, &mode, NULL, &pairable);
+	btd_adapter_get_mode(adapter, &mode, NULL, NULL);
 	if (mode == MODE_OFF) {
 		mgmt_set_powered(index, FALSE);
 		return;
 	}
 
-	discoverable = (mode == MODE_DISCOVERABLE);
-
-	if (info->discoverable != discoverable)
-		mgmt_set_discoverable(index, discoverable);
-
-	if (info->pairable != pairable)
-		mgmt_set_pairable(index, pairable);
-
 	if (info->enabled)
-		btd_adapter_start(adapter);
+		mgmt_update_powered(index, TRUE);
+	else
+		mgmt_set_powered(index, TRUE);
 
 	btd_adapter_unref(adapter);
 }
 
+static void set_powered_complete(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *rp = buf;
+	uint16_t index;
+
+	if (len < sizeof(*rp)) {
+		error("Too small set powered complete event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&rp->index));
+
+	DBG("hci%d powered %u", index, rp->val);
+
+	mgmt_update_powered(index, rp->val);
+}
+
+static void set_discoverable_complete(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *rp = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+	uint8_t mode;
+
+	if (len < sizeof(*rp)) {
+		error("Too small set discoverable complete event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&rp->index));
+
+	DBG("hci%d discoverable %u", index, rp->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in discoverable complete", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->discoverable = rp->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
+		return;
+
+	/* set_discoverable will always also change page scanning */
+	mode = SCAN_PAGE;
+
+	if (info->discoverable)
+		mode |= SCAN_INQUIRY;
+
+	adapter_mode_changed(adapter, mode);
+}
+
+static void set_connectable_complete(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *rp = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+
+	if (len < sizeof(*rp)) {
+		error("Too small set connectable complete event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&rp->index));
+
+	DBG("hci%d connectable %u", index, rp->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in connectable complete", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->connectable = rp->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (adapter)
+		adapter_mode_changed(adapter, rp->val ? SCAN_PAGE : 0);
+}
+
+static void set_pairable_complete(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *rp = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+
+	if (len < sizeof(*rp)) {
+		error("Too small set pairable complete event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&rp->index));
+
+	DBG("hci%d pairable %u", index, rp->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in pairable complete", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->pairable = rp->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
+		return;
+
+	btd_adapter_pairable_changed(adapter, info->pairable);
+}
+
 static void mgmt_cmd_complete(int sk, void *buf, size_t len)
 {
 	struct mgmt_ev_cmd_complete *ev = buf;
@@ -360,6 +683,24 @@ static void mgmt_cmd_complete(int sk, void *buf, size_t len)
 	case MGMT_OP_READ_INFO:
 		read_info_complete(sk, ev->data, len - sizeof(*ev));
 		break;
+	case MGMT_OP_SET_POWERED:
+		set_powered_complete(sk, ev->data, len - sizeof(*ev));
+		break;
+	case MGMT_OP_SET_DISCOVERABLE:
+		set_discoverable_complete(sk, ev->data, len - sizeof(*ev));
+		break;
+	case MGMT_OP_SET_CONNECTABLE:
+		set_connectable_complete(sk, ev->data, len - sizeof(*ev));
+		break;
+	case MGMT_OP_SET_PAIRABLE:
+		set_pairable_complete(sk, ev->data, len - sizeof(*ev));
+		break;
+	case MGMT_OP_ADD_UUID:
+		DBG("add_uuid complete");
+		break;
+	case MGMT_OP_REMOVE_UUID:
+		DBG("remove_uuid complete");
+		break;
 	default:
 		error("Unknown command complete for opcode %u", opcode);
 		break;
@@ -457,6 +798,15 @@ static gboolean mgmt_event(GIOChannel *io, GIOCondition cond, gpointer user_data
 	case MGMT_EV_POWERED:
 		mgmt_powered(sk, buf + MGMT_HDR_SIZE, len);
 		break;
+	case MGMT_EV_DISCOVERABLE:
+		mgmt_discoverable(sk, buf + MGMT_HDR_SIZE, len);
+		break;
+	case MGMT_EV_CONNECTABLE:
+		mgmt_connectable(sk, buf + MGMT_HDR_SIZE, len);
+		break;
+	case MGMT_EV_PAIRABLE:
+		mgmt_pairable(sk, buf + MGMT_HDR_SIZE, len);
+		break;
 	default:
 		error("Unknown Management opcode %u", opcode);
 		break;
@@ -488,7 +838,7 @@ static int mgmt_setup(void)
 	}
 
 	memset(&hdr, 0, sizeof(hdr));
-	hdr.opcode = MGMT_OP_READ_VERSION;
+	hdr.opcode = htobs(MGMT_OP_READ_VERSION);
 	if (write(dd, &hdr, sizeof(hdr)) < 0) {
 		err = -errno;
 		goto fail;
@@ -527,12 +877,6 @@ static void mgmt_cleanup(void)
 	}
 }
 
-static int mgmt_set_connectable(int index, gboolean connectable)
-{
-	DBG("index %d connectable %d", index, connectable);
-	return -ENOSYS;
-}
-
 static int mgmt_set_dev_class(int index, uint8_t major, uint8_t minor)
 {
 	DBG("index %d major %u minor %u", index, major, minor);
@@ -779,10 +1123,16 @@ static int mgmt_set_did(int index, uint16_t vendor, uint16_t product,
 	return -ENOSYS;
 }
 
-static int mgmt_services_updated(int index)
+static int mgmt_add_uuid(int index, uuid_t *uuid)
 {
 	DBG("index %d", index);
-	return -ENOSYS;
+	return mgmt_uuid_op(index, MGMT_OP_ADD_UUID, uuid);
+}
+
+static int mgmt_remove_uuid(int index, uuid_t *uuid)
+{
+	DBG("index %d", index);
+	return mgmt_uuid_op(index, MGMT_OP_REMOVE_UUID, uuid);
 }
 
 static int mgmt_disable_cod_cache(int index)
@@ -808,7 +1158,6 @@ static struct btd_adapter_ops mgmt_ops = {
 	.setup = mgmt_setup,
 	.cleanup = mgmt_cleanup,
 	.set_powered = mgmt_set_powered,
-	.set_connectable = mgmt_set_connectable,
 	.set_discoverable = mgmt_set_discoverable,
 	.set_pairable = mgmt_set_pairable,
 	.set_limited_discoverable = mgmt_set_limited_discoverable,
@@ -839,7 +1188,8 @@ static struct btd_adapter_ops mgmt_ops = {
 	.enable_le = mgmt_enable_le,
 	.encrypt_link = mgmt_encrypt_link,
 	.set_did = mgmt_set_did,
-	.services_updated = mgmt_services_updated,
+	.add_uuid = mgmt_add_uuid,
+	.remove_uuid = mgmt_remove_uuid,
 	.disable_cod_cache = mgmt_disable_cod_cache,
 	.restore_powered = mgmt_restore_powered,
 	.load_keys = mgmt_load_keys,
